/* 
 * VEX, a visual editor for XML
 *
 * Copyright (c) 2005 John Krasnay
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
package net.sf.vex.layout;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import net.sf.vex.core.Caret;
import net.sf.vex.core.Color;
import net.sf.vex.core.ColorResource;
import net.sf.vex.core.FontMetrics;
import net.sf.vex.core.FontResource;
import net.sf.vex.core.FontSpec;
import net.sf.vex.core.Graphics;
import net.sf.vex.core.Insets;
import net.sf.vex.core.IntRange;
import net.sf.vex.core.Rectangle;
import net.sf.vex.css.CSS;
import net.sf.vex.css.StyleSheet;
import net.sf.vex.css.Styles;
import net.sf.vex.dom.IVexDocument;
import net.sf.vex.dom.IVexElement;
import net.sf.vex.dom.Position;

/**
 * Base class of block boxes that can contain other block boxes. This class
 * implements the layout method and various navigation methods. Subclasses must
 * implement the createChildren method.
 * 
 * Subclasses can be anonymous or non-anonymous (i.e. generated by an element).
 * Since the vast majority of instances will be non-anonymous, this class can
 * manage the element and top and bottom margins without too much inefficiency.
 * 
 * <p>
 * Subclasses that can be anonymous must override the getStartPosition and
 * getEndPosition classes to return the range covered by the box.
 * </p>
 */
public abstract class AbstractBlockBox extends AbstractBox implements BlockBox {

	private static Color markerColor = new Color(100, 100, 160);

	/**
	 * Class constructor for non-anonymous boxes.
	 * 
	 * @param context
	 *            LayoutContext being used.
	 * @param parent
	 *            Parent box.
	 * @param element
	 *            Element associated with this box. anonymous box.
	 */
	public AbstractBlockBox(LayoutContext context, BlockBox parent,
			IVexElement element) {

		this.parent = parent;
		this.element = element;

		int parentWidth = parent.getWidth();

		if (element != null) {
			Styles styles = context.getStyleSheet().getStyles(element);
			this.marginTop = styles.getMarginTop().get(parentWidth);
			this.marginBottom = styles.getMarginBottom().get(parentWidth);
		}

	}

	/**
	 * Class constructor for anonymous boxes.
	 * 
	 * @param context
	 *            LayoutContext to use.
	 * @param parent
	 *            Parent box.
	 * @param startOffset
	 *            Start of the range covered by the box.
	 * @param endOffset
	 *            End of the range covered by the box.
	 */
	public AbstractBlockBox(LayoutContext context, BlockBox parent,
			int startOffset, int endOffset) {
		this.parent = parent;
		this.marginTop = 0;
		this.marginBottom = 0;

		IVexDocument doc = context.getDocument();
		this.startPosition = doc.createPosition(startOffset);
		this.endPosition = doc.createPosition(endOffset);
	}

	/**
	 * Walks the box tree and returns the nearest enclosing element.
	 */
	protected IVexElement findContainingElement() {
		BlockBox box = this;
		IVexElement element = box.getElement();
		while (element == null) {
			box = box.getParent();
			element = box.getElement();
		}
		return element;
	}

	/**
	 * Returns this box's children as an array of BlockBoxes.
	 */
	protected BlockBox[] getBlockChildren() {
		return (BlockBox[]) this.getChildren();
	}

	public Caret getCaret(LayoutContext context, int offset) {

		// If we haven't yet laid out this block, estimate the caret.
		if (this.getLayoutState() != LAYOUT_OK) {
			int relative = offset - this.getStartOffset();
			int size = this.getEndOffset() - this.getStartOffset();
			int y = 0;
			if (size > 0) {
				y = this.getHeight() * (relative / size);
			}
			return new HCaret(0, y, this.getHCaretWidth());
		}

		int y;

		Box[] contentChildren = this.getContentChildren();
		for (int i = 0; i < contentChildren.length; i++) {

			if (offset < contentChildren[i].getStartOffset()) {
				if (i > 0) {
					y = (contentChildren[i - 1].getY() + contentChildren[i - 1].getHeight() + contentChildren[i]
							.getY()) / 2;
				} else {

					Box firstChild = getFirstChild();
					if (firstChild != null
							&& firstChild instanceof BlockElementWidget) {
						y = firstChild.getHeight();
					} else {
						y = 0;
					}
				}
				return new HCaret(0, y, this.getHCaretWidth());
			}

			if (offset >= contentChildren[i].getStartOffset()
					&& offset <= contentChildren[i].getEndOffset()) {

				Caret caret = contentChildren[i].getCaret(context, offset);
				caret.translate(contentChildren[i].getX(), contentChildren[i].getY());
				return caret;
			}
		}

		if (this.hasChildren()) {
			y = this.getHeight();
		} else {
			y = this.getHeight() / 2;
		}

		return new HCaret(0, y, this.getHCaretWidth());
	}

	public Box[] getChildren() {
		return this.children;
	}

	/**
	 * Return an array of children that contain content.
	 */
	protected BlockBox[] getContentChildren() {
		Box[] children = this.getChildren();
		List contentChildren = new ArrayList(children.length);
		for (int i = 0; i < children.length; i++) {
			if (children[i].hasContent()) {
				contentChildren.add(children[i]);
			}
		}
		return (BlockBox[]) contentChildren
				.toArray(new BlockBox[contentChildren.size()]);
	}

	public IVexElement getElement() {
		return this.element;
	}

	public int getEndOffset() {
		IVexElement element = this.getElement();
		if (element != null) {
			return element.getEndOffset();
		} else if (this.getEndPosition() != null) {
			return this.getEndPosition().getOffset();
		} else {
			throw new IllegalStateException();
		}
	}

	/**
	 * Returns the estimated size of the box, based on the the current font size
	 * and the number of characters covered by the box. This is a utility method
	 * that can be used in implementation of setInitialSize. It assumes the
	 * width of the box has already been correctly set.
	 * 
	 * @param context
	 *            LayoutContext to use.
	 */
	protected int getEstimatedHeight(LayoutContext context) {

		IVexElement element = this.findContainingElement();
		Styles styles = context.getStyleSheet().getStyles(element);
		int charCount = this.getEndOffset() - this.getStartOffset();

		float fontSize = styles.getFontSize();
		float lineHeight = styles.getLineHeight();
		float estHeight = lineHeight * fontSize * 0.6f * charCount
				/ this.getWidth();

		return Math.round(Math.max(estHeight, lineHeight));
	}

	public LineBox getFirstLine() {
		if (this.hasChildren()) {
			BlockBox firstChild = (BlockBox) this.getChildren()[0];
			return firstChild.getFirstLine();
		} else {
			return null;
		}
	}

	/**
	 * Returns the width of the horizontal caret. This is overridden by TableBox
	 * to return a caret that is the full width of the table.
	 */
	protected int getHCaretWidth() {
		return H_CARET_LENGTH;
	}

	public LineBox getLastLine() {
		if (this.hasChildren()) {
			BlockBox lastChild = (BlockBox) this.getChildren()[this
					.getChildren().length - 1];
			return lastChild.getLastLine();
		} else {
			return null;
		}
	}

	/**
	 * Returns the layout state of this box.
	 */
	protected byte getLayoutState() {
		return this.layoutState;
	}

	public int getLineEndOffset(int offset) {
		BlockBox[] children = this.getContentChildren();
		for (int i = 0; i < children.length; i++) {
			if (children[i].containsOffset(offset)) {
				return children[i].getLineEndOffset(offset);
			}
		}
		return offset;
	}

	public int getLineStartOffset(int offset) {
		BlockBox[] children = this.getContentChildren();
		for (int i = 0; i < children.length; i++) {
			if (children[i].containsOffset(offset)) {
				return children[i].getLineStartOffset(offset);
			}
		}
		return offset;
	}

	public int getMarginBottom() {
		return this.marginBottom;
	}

	public int getMarginTop() {
		return this.marginTop;
	}

	public int getNextLineOffset(LayoutContext context, int offset, int x) {

		//
		// This algorithm works when this block owns the offsets between
		// its children.
		//

		if (offset == this.getEndOffset()) {
			return -1;
		}

		BlockBox[] children = this.getContentChildren();

		if (offset < this.getStartOffset() && children.length > 0
				&& children[0].getStartOffset() > this.getStartOffset()) {
			//
			// If there's an offset before the first child, put the caret there.
			//
			return this.getStartOffset();
		}

		for (int i = 0; i < children.length; i++) {
			BlockBox child = children[i];
			if (offset <= child.getEndOffset()) {
				int newOffset = child.getNextLineOffset(context, offset, x
						- child.getX());
				if (newOffset < 0 /* && i < children.length-1 */) {
					return child.getEndOffset() + 1;
				} else {
					return newOffset;
				}
			}
		}

		return this.getEndOffset();
	}

	public BlockBox getParent() {
		return this.parent;
	}

	public int getPreviousLineOffset(LayoutContext context, int offset, int x) {

		if (offset == this.getStartOffset()) {
			return -1;
		}

		BlockBox[] children = this.getContentChildren();

		if (offset > this.getEndOffset()
				&& children.length > 0
				&& children[children.length - 1].getEndOffset() < this
						.getEndOffset()) {
			//
			// If there's an offset after the last child, put the caret there.
			//
			return this.getEndOffset();
		}

		for (int i = children.length; i > 0; i--) {
			BlockBox child = children[i - 1];
			if (offset >= child.getStartOffset()) {
				int newOffset = child.getPreviousLineOffset(context, offset, x
						- child.getX());
				if (newOffset < 0 && i > 0) {
					return child.getStartOffset() - 1;
				} else {
					return newOffset;
				}
			}
		}

		return this.getStartOffset();
	}

	public int getStartOffset() {
		IVexElement element = this.getElement();
		if (element != null) {
			return element.getStartOffset() + 1;
		} else if (this.getStartPosition() != null) {
			return this.getStartPosition().getOffset();
		} else {
			throw new IllegalStateException();
		}
	}

	public boolean hasContent() {
		return true;
	}

	public void invalidate(boolean direct) {

		if (direct) {
			this.layoutState = LAYOUT_REDO;
		} else {
			this.layoutState = LAYOUT_PROPAGATE;
		}

		if (this.getParent() instanceof AbstractBlockBox) {
			((AbstractBlockBox) this.getParent()).invalidate(false);
		}
	}

	public boolean isAnonymous() {
		return this.getElement() == null;
	}

	/**
	 * Call the given callback for each child matching one of the given display
	 * styles. Any nodes that do not match one of the given display types cause
	 * the onRange callback to be called, with a range covering all such
	 * contiguous nodes.
	 * 
	 * @param styleSheet
	 *            StyleSheet from which to determine display styles.
	 * @param displayStyles
	 *            Display types to be explicitly recognized.
	 * @param callback
	 *            DisplayStyleCallback through which the caller is notified of
	 *            matching elements and non-matching ranges.
	 */
	protected void iterateChildrenByDisplayStyle(StyleSheet styleSheet,
			Set displayStyles, ElementOrRangeCallback callback) {
		LayoutUtils.iterateChildrenByDisplayStyle(styleSheet, displayStyles,
				this.findContainingElement(), this.getStartOffset(), this
						.getEndOffset(), callback);
	}

	public void paint(LayoutContext context, int x, int y, Rectangle area) {

		if (this.skipPaint(context, x, y)) {
			return;
		}

		this.drawBox(context, x, y, this.getParent().getWidth(), true);

		this.paintChildren(context, x, y, area);

		// this.paintSelectionFrame(context, x, y, true);
	}

	/**
	 * Default implementation. Width is calculated as the parent's width minus
	 * this box's insets. Height is calculated by getEstimatedHeight.
	 */
	public void setInitialSize(LayoutContext context) {
		int parentWidth = this.getParent().getWidth();
		Insets insets = this.getInsets(context, parentWidth);
		this.setWidth(Math.max(0, parentWidth - insets.getLeft()
				- insets.getRight()));
		this.setHeight(this.getEstimatedHeight(context));
	}

	public int viewToModel(LayoutContext context, int x, int y) {

		Box[] children = this.getChildren();

		if (children == null) {
			int charCount = this.getEndOffset() - this.getStartOffset() - 1;
			if (charCount == 0 || this.getHeight() == 0) {
				return this.getEndOffset();
			} else {
				return this.getStartOffset() + charCount * y / this.getHeight();
			}
		} else {

			// Linear Search
			for (int i = 0; i < children.length; i++) {
				Box child = children[i];
				if (!child.hasContent()) {
					continue;
				}
				if (y < child.getY()) {
					return child.getStartOffset() - 1;
				} else if (y < child.getY() + child.getHeight()) {
					return child.viewToModel(context, x - child.getX(), y
							- child.getY());
				}
			}

			// Binary search to speed this up

			// if (y < 0 ) return getStartOffset();
			//			
			// System.out.println();
			// System.out.println("*  VIEW_TO_MODEL: BinSearch");
			// System.out.println("*  Searching: " + y);
			//
			// int first = 0;
			// int last = children.length -1 ;
			// int middle;
			// int firstY;
			// int lastY;
			// int middleY;
			//
			//
			// while (first <= last ) {
			//
			//
			//
			// middle = first + ((last - first) / 2);
			//				
			// firstY = children[first].getY();
			// middleY = middle <= 0 ? 0 : children[middle].getY();
			// lastY = children[last].getY();
			//
			//
			//				
			//				
			// System.out.println("*  First: " + children[first].getY());
			// System.out.println("*  Last: " + children[last].getY());
			// System.out.println("*  Middle: " + children[middle].getY());
			//
			//
			// if (y >= middleY) {
			// if (y < (middle==children.length-1 ? children[middle].getY()+
			// children[middle].getHeight() : children[middle +1].getY())) {
			//
			//
			// return children [middle].viewToModel(context,
			// x - children[middle] .getX(), Math.max(0,y-
			// children[middle].getY()));
			// }else {
			// first = middle + 1;
			// }
			// } else if (y < middleY) {
			// last = middle -1;
			// }
			//
			// }
		}

		return this.getEndOffset();
	}

	// ===================================================== PRIVATE

	private BlockBox parent;

	protected Box[] children;

	/**
	 * Paint a frame around the block element. Can be used for indicating a
	 * selection or to display the underlaying document structure.
	 * 
	 * @param context
	 *            LayoutContext to use.
	 * @param x
	 *            x-coordinate at which to draw
	 * @param y
	 *            y-coordinate at which to draw.
	 * 
	 * @param width
	 *            width of the surrounding element
	 */
	protected void paintSurroundingFrame(LayoutContext context, int x, int y, int width, ColorResource foreground,
			ColorResource background) {

		Styles styles = context.getStyleSheet().getStyles(element);

		int plusLeft = styles.getBorderLeftWidth()
				+ styles.getPaddingLeft().get(width);
		int plusTop = styles.getBorderTopWidth()
				+ styles.getPaddingTop().get(width);
		int plusRight = styles.getBorderRightWidth()
				+ styles.getPaddingRight().get(width);
		int plusBottom = styles.getBorderBottomWidth()
				+ styles.getPaddingBottom().get(width);

		Graphics g = context.getGraphics();

		FontResource font = g.createFont(new FontSpec(
				new String[] { "sans-serif" }, FontSpec.PLAIN, 10));
		FontResource oldFont = g.setFont(font);
		FontMetrics fm = g.getFontMetrics();
		ColorResource oldColor = g.setColor(background);
		g.setLineStyle(Graphics.LINE_SOLID);
		g.setLineWidth(1);
		int tabWidth = g.stringWidth(this.getElement().getName())
				+ fm.getLeading();
		int tabHeight = fm.getHeight();
		int tabX = x + this.getWidth() - tabWidth + plusRight;
		int tabY = y - plusTop; // To go to bottom: + this.getHeight() -
		// tabHeight;
		g.drawRect(x - plusLeft, y - plusTop, this.getWidth() + plusLeft
				+ plusRight, this.getHeight() + plusTop + plusBottom);
		// Prints the element's name in the upper right corner
		// now obsolete
		// g.fillRect(tabX, tabY, tabWidth, tabHeight);
		// g.setColor(foreground);
		// g.drawString(this.getElement().getName(), tabX + fm.getLeading() / 2,
		// tabY);

		g.setColor(oldColor);
		g.setFont(oldFont);
		font.dispose();

	}

	/**
	 * Paint a frame to mark and label any block elements.
	 * 
	 * @param context
	 *            LayoutContext to use.
	 * @param x
	 *            x-coordinate at which to draw
	 * @param y
	 *            y-coordinate at which to draw.
	 * @param selected
	 */
	protected void paintMarkerFrame(LayoutContext context, int x, int y,
			int width) {

		ColorResource foreground;
		ColorResource background;

		AnnotationRenderingInfo annotationInfo = VexAnnotationTracker.getRenderingInfo(context, getElement());
		if (annotationInfo != null && annotationInfo.isShownInText())
			background = context.getGraphics().createColor(annotationInfo.getColor());
		else
			background = context.getGraphics().createColor(markerColor);

		foreground = context.getGraphics()
				.createColor(new Color(255, 255, 255));

		paintSurroundingFrame(context, x, y, width, foreground, background);

		foreground.dispose();
		background.dispose();
	}

	/**
	 * Paint a frame that indicates a block element box has been selected. This
	 * frame also displays the elements name. Note that you also need to call
	 * paintSelectionFill().
	 * 
	 * @param context
	 *            LayoutContext to use.
	 * @param x
	 *            x-coordinate at which to draw
	 * @param y
	 *            y-coordinate at which to draw.
	 * 
	 */
	protected void paintSelectionFrame(LayoutContext context, int x, int y,
			int width) {

		ColorResource background = context.getGraphics().getSystemColor(
				ColorResource.SELECTION_BACKGROUND);

		ColorResource foreground = context.getGraphics().getSystemColor(
				ColorResource.SELECTION_FOREGROUND);

		paintSurroundingFrame(context, x, y, width, background, foreground);

	}

	/**
	 * Simply paints a rectangle filled with the selection background color.
	 * After this, paint the elements children. Then paint decorations with
	 * paintSelectionFrame().
	 * 
	 * @param context
	 *            LayoutContext to use.
	 * @param x
	 *            x-coordinate at which to draw
	 * @param y
	 *            y-coordinate at which to draw.
	 * 
	 */

	protected void paintSelectionFill(LayoutContext context, int x, int y) {

		ColorResource background = context.getGraphics().getSystemColor(
				ColorResource.SELECTION_BACKGROUND);

		Graphics g = context.getGraphics();

		ColorResource oldColor = g.setColor(background);

		g.fillRect(x, y, this.getWidth(), this.getHeight());

		g.setColor(oldColor);

	}

	/** Layout is OK */
	public static final byte LAYOUT_OK = 0;

	/** My layout is OK, but one of my children needs to be laid out */
	public static final byte LAYOUT_PROPAGATE = 1;

	/** I need to be laid out */
	public static final byte LAYOUT_REDO = 2;

	private byte layoutState = LAYOUT_REDO;

	public IntRange layout(LayoutContext context, int top, int bottom) {

		int repaintStart = Integer.MAX_VALUE;
		int repaintEnd = 0;
		boolean repaintToBottom = false;

		Insets ins = this.getInsets(context, -1);

		int originalHeight = this.getHeight() + ins.getTop() + ins.getBottom();

		final String name = this.getElement() == null ? this.toString() : getElement().getName();
		context.subTask("Laying out {0}", name);
		if (this.layoutState == LAYOUT_REDO) {


			List childList = this.createChildren(context);
			this.children = (BlockBox[]) childList
					.toArray(new BlockBox[childList.size()]);

			// Even though we position children after layout, we have to
			// do a preliminary positioning here so we now which ones
			// overlap our layout band
			for (int i = 0; i < this.children.length; i++) {
				BlockBox child = (BlockBox) this.children[i];
				child.setInitialSize(context);
			}
			this.positionChildren(context);

			// repaint everything
			repaintToBottom = true;
			repaintStart = 0;
		}

		// If element hasn't changed, don't do anything.
		// if (this.layoutState != LAYOUT_OK) {
		Box[] children = this.getChildren();
		for (int i = 0; i < children.length; i++) {
			if (children[i] instanceof BlockBox) {
				BlockBox child = (BlockBox) children[i];
				Insets childIns = Insets.ZERO_INSETS;
				if (child instanceof BlockElementBox)
					childIns = child.getInsets(context, this.getWidth());
				if (top <= child.getY() + child.getHeight() + childIns.getTop()
						+ childIns.getBottom()) {
					if (bottom >= child.getY()) {

						IntRange repaintRange = child.layout(context, top
								- child.getY(), bottom - child.getY());
						if (repaintRange != null) {
							repaintStart = Math.min(repaintStart, repaintRange
									.getStart()
									+ child.getY());
							repaintEnd = Math.max(repaintEnd, repaintRange
									.getEnd()
									+ child.getY());
						}
					}
					// children are below the viewed area from here, we can quit
					// else break;
				}
			}
		}

		int childRepaintStart = this.positionChildren(context);
		if (childRepaintStart != -1) {
			repaintToBottom = true;
			repaintStart = Math.min(repaintStart, childRepaintStart);
		}

		this.layoutState = LAYOUT_OK;
		// }

		if (repaintToBottom) {
			ins = Insets.ZERO_INSETS;
			if (this instanceof BlockElementBox)
				ins = this.getInsets(context, -1);

			repaintEnd = Math.max(originalHeight, this.getHeight()
					+ ins.getTop() + ins.getBottom());
		}

		if (repaintStart < repaintEnd) {
			return new IntRange(repaintStart, repaintEnd);
		} else {
			return null;
		}
	}

	protected abstract List createChildren(LayoutContext context);

	/**
	 * Creates a list of block boxes for a given document range. beforeInlines
	 * and afterInlines are prepended/appended to the first/last block child,
	 * and each may be null.
	 */
	protected List createBlockBoxes(LayoutContext context, int startOffset,
			int endOffset, int width, List beforeInlines, List afterInlines) {

		List blockBoxes = new ArrayList();
		List pendingInlines = new ArrayList();

		if (beforeInlines != null) {
			pendingInlines.addAll(beforeInlines);
		}

		IVexElement element = context.getDocument().findCommonElement(
				startOffset, endOffset);

		if (startOffset == endOffset) {
			int relOffset = startOffset - element.getStartOffset();
			pendingInlines.add(new PlaceholderBox(context, element, relOffset));
		} else {

			BlockInlineIterator iter = new BlockInlineIterator(context,
					element, startOffset, endOffset);

			while (true) {

				Object next = iter.next();

				if (next == null) {
					break;
				}

				if (next instanceof IntRange) {

					IntRange range = (IntRange) next;

					InlineElementBox.InlineBoxes inlineBoxes = InlineElementBox
							.createInlineBoxes(context, element, range
									.getStart(), range.getEnd());
					pendingInlines.addAll(inlineBoxes.boxes);
					pendingInlines.add(new PlaceholderBox(context, element,
							range.getEnd() - element.getStartOffset()));

				} else {

					if (pendingInlines.size() > 0) {
						blockBoxes.add(ParagraphBox.create(context, element,
								pendingInlines, width));
						pendingInlines.clear();
					}

					if (isTableChild(context, next)) {

						// Consume continguous table children and create an
						// anonymous table.

						int tableStartOffset = ((IVexElement) next)
								.getStartOffset();
						int tableEndOffset = -1; // dummy to hide warning
						while (isTableChild(context, next)) {
							tableEndOffset = ((IVexElement) next)
									.getEndOffset() + 1;
							next = iter.next();
						}

						// add anonymous table
						blockBoxes.add(new TableBox(context, this,
								tableStartOffset, tableEndOffset));

						if (next == null) {
							break;
						} else {
							iter.push(next);
						}

					} else { // next is a block box element
						IVexElement blockElement = (IVexElement) next;
						blockBoxes.add(context.getBoxFactory().createBox(
								context, blockElement, this, width));
					}
				}
			}
		}

		if (afterInlines != null) {
			pendingInlines.addAll(afterInlines);
		}

		if (pendingInlines.size() > 0) {
			blockBoxes.add(ParagraphBox.create(context, element,
					pendingInlines, width));
			pendingInlines.clear();
		}

		return blockBoxes;
	}

	private class BlockInlineIterator {

		public BlockInlineIterator(LayoutContext context, IVexElement element,
				int startOffset, int endOffset) {
			this.context = context;
			this.element = element;
			this.startOffset = startOffset;
			this.endOffset = endOffset;
		}

		/**
		 * Returns the next block element or inline range, or null if we're at
		 * the end.
		 */
		public Object next() {
			if (this.pushStack.size() > 0) {
				return this.pushStack.removeLast();
			} else if (startOffset == endOffset) {
				return null;
			} else {
				IVexElement blockElement = findNextBlockElement(this.context,
						this.element, startOffset, endOffset);
				if (blockElement == null) {
					if (startOffset < endOffset) {
						IntRange result = new IntRange(startOffset, endOffset);
						startOffset = endOffset;
						return result;
					} else {
						return null;
					}
				} else if (blockElement.getStartOffset() > startOffset) {
					this.pushStack.addLast(blockElement);
					IntRange result = new IntRange(startOffset, blockElement
							.getStartOffset());
					startOffset = blockElement.getEndOffset() + 1;
					return result;
				} else {
					startOffset = blockElement.getEndOffset() + 1;
					return blockElement;
				}
			}
		}

		public Object peek() {
			if (this.pushStack.size() == 0) {
				Object next = next();
				if (next == null) {
					return null;
				} else {
					push(next);
				}
			}
			return pushStack.getLast();
		}

		public void push(Object pushed) {
			this.pushStack.addLast(pushed);
		}

		private LayoutContext context;

		private IVexElement element;

		private int startOffset;

		private int endOffset;

		private LinkedList pushStack = new LinkedList();
	}

	protected boolean hasChildren() {
		return this.getChildren() != null && this.getChildren().length > 0;
	}

	/**
	 * Positions the children of this box. Vertical margins are collapsed here.
	 * Returns the vertical offset of the top of the first child to move, or -1
	 * if not children were actually moved.
	 */
	protected int positionChildren(LayoutContext context) {

		int childY = 0; // TODO tabWidth
		int prevMargin = 0;

		BlockBox[] children = this.getBlockChildren();
		int repaintStart = -1;

		Styles styles = null;

		if (!this.isAnonymous()) {
			styles = context.getStyleSheet().getStyles(this.getElement());
		}

		if (styles != null && children.length > 0) {
			if (styles.getBorderTopWidth()
					+ styles.getPaddingTop().get(this.getWidth()) == 0) {
				// first child's top margin collapses into ours
				this.marginTop = Math.max(this.marginTop, children[0]
						.getMarginTop());
				childY -= children[0].getMarginTop();
			}
		}

		for (int i = 0; i < children.length; i++) {

			Insets insets = children[i].getInsets(context, this.getWidth());

			// if (!children[i].isAnonymous()) {
			//				
			// Styles childStyles =
			// context.getStyleSheet().getStyles(children[i].getElement());
			//				
			// insets = new
			// Insets(childStyles.getPaddingTop().get(children[i].getWidth()) +
			// childStyles.getBorderTopWidth() +
			// childStyles.getMarginTop().get(children[i].getWidth()),
			// childStyles.getPaddingLeft().get(children[i].getWidth()) +
			// childStyles.getBorderLeftWidth() +
			// childStyles.getMarginLeft().get(children[i].getWidth()),
			// childStyles.getPaddingBottom().get(children[i].getWidth()) +
			// childStyles.getBorderBottomWidth() +
			// childStyles.getMarginBottom().get(children[i].getWidth()),
			// childStyles.getPaddingRight().get(children[i].getWidth()) +
			// childStyles.getBorderRightWidth() +
			// childStyles.getMarginRight().get(children[i].getWidth()));
			// }

			childY += insets.getTop();

			if (i > 0) {
				childY -= Math.min(prevMargin, children[i].getMarginTop());
			}

			if (repaintStart == -1 && children[i].getY() != childY) {
				repaintStart = Math.min(children[i].getY(), childY);
			}

			children[i].setX(insets.getLeft());
			children[i].setY(childY);

			childY += children[i].getHeight() + insets.getBottom();
			prevMargin = children[i].getMarginBottom();
		}

		if (styles != null && children.length > 0) {
			if (styles.getBorderBottomWidth()
					+ styles.getPaddingBottom().get(this.getWidth()) == 0) {
				// last child's bottom margin collapses into ours
				this.marginBottom = Math.max(this.marginBottom, prevMargin);
				childY -= prevMargin;
			}
		}

		this.setHeight(childY);

		return repaintStart;
	}

	/**
	 * Sets the layout state of the box.
	 * 
	 * @param layoutState
	 *            One of the LAYOUT_* constants
	 */
	protected void setLayoutState(byte layoutState) {
		this.layoutState = layoutState;
	}

	// ========================================================= PRIVATE
	/** The length, in pixels, of the horizontal caret between block boxes */
	private static final int H_CARET_LENGTH = 20;

	/**
	 * Element with which we are associated. For anonymous boxes, this is null.
	 */
	private IVexElement element;

	/*
	 * We cache the top and bottom margins, since they may be affected by our
	 * children.
	 */
	private int marginTop;

	private int marginBottom;

	/**
	 * Start position of an anonymous box. For non-anonymous boxes, this is
	 * null.
	 */
	protected Position startPosition;

	/**
	 * End position of an anonymous box. For non-anonymous boxes, this is null.
	 */
	protected Position endPosition;

	/**
	 * Searches for the next block-formatted child. <em> NOTE </em> that this
	 * will also return children of children etc.
	 * 
	 * @param context
	 *            LayoutContext to use.
	 * @param element
	 *            Element within which to search.
	 * @param startOffset
	 *            The offset at which to start the search.
	 * @param endOffset
	 *            The offset at which to end the search.
	 */
	private static IVexElement findNextBlockElement(LayoutContext context,
			IVexElement element, int startOffset, int endOffset) {

		IVexElement[] children = element.getChildElements();
		for (int i = 0; i < children.length; i++) {
			IVexElement child = children[i];
			if (child.getEndOffset() < startOffset) {
				continue;
			} else if (child.getStartOffset() >= endOffset) {
				break;
			} else {
				Styles styles = context.getStyleSheet().getStyles(child);
				if (!styles.getDisplay().equalsIgnoreCase(CSS.INLINE)) { // TODO
					// do
					// proper block
					// display
					// determination
					return child;
				} else {
					continue;

				}
			}
		}

		return null;
	}

	/**
	 * Return the end position of an anonymous box. The default implementation
	 * returns null.
	 */
	private Position getEndPosition() {
		return this.endPosition;
	}

	/**
	 * Return the start position of an anonymous box. The default implementation
	 * returns null.
	 */
	private Position getStartPosition() {
		return this.startPosition;
	}

	private boolean isTableChild(LayoutContext context, Object rangeOrElement) {
		if (rangeOrElement != null && rangeOrElement instanceof IVexElement) {
			return LayoutUtils.isTableChild(context.getStyleSheet(),
					(IVexElement) rangeOrElement);
		} else {
			return false;
		}
	}

	@Override
	public Insets getInsets(LayoutContext context, int containerWidth) {

		Insets ins = new Insets(marginTop, 0, marginBottom, 0);

		return ins;
	}

	public void setParent(BlockBox parent) {
		this.parent = parent;
	}

	public static void setMarkerColor(Color markerColor) {
		AbstractBlockBox.markerColor = markerColor;
	}

	/**
	 * Returns the root box
	 */
	public RootBox findRoot() {
		BlockBox b = this;
		while (!(b instanceof RootBox)) {
			b = b.getParent();
		}

		return (RootBox) b;
	}

	/**
	 * Return the first child that is not a BinTree
	 * @return
	 */
	public Box getFirstChild() {

		
		if (children.length != 0 )
			if (children[0] instanceof BinTree)
				return ((AbstractBlockBox)children[0]).getFirstChild();
			else 
				return children[0];
		else
			return null;
				

	}
}
